@@ -1,7 +1,7 @@
 /*
  *  npruntime.c - Scripting plugins support
  *
- *  nspluginwrapper (C) 2005-2007 Gwenole Beauchesne
+ *  nspluginwrapper (C) 2005-2009 Gwenole Beauchesne
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -13,9 +13,9 @@
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *  You should have received a copy of the GNU General Public License along
+ *  with this program; if not, write to the Free Software Foundation, Inc.,
+ *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 #include "sysdeps.h"
@@ -23,15 +23,8 @@
 #include <assert.h>
 #include <glib.h> /* <glib/ghash.h> */
 #include "utils.h"
-#include "rpc.h"
-#include "npw-rpc.h"
-
-#define XP_UNIX 1
-#define MOZ_X11 1
-#include <npapi.h>
-#include <npupp.h>
-#include <npruntime.h>
-#include "npruntime-impl.h"
+#include "npw-common.h"
+#include "npw-malloc.h"
 
 #define DEBUG 1
 #include "debug.h"
@@ -40,28 +33,58 @@
 // Defined in npw-{wrapper,viewer}.c
 extern rpc_connection_t *g_rpc_connection attribute_hidden;
 
+// Defined in npw-viewer.c
+#if USE_PID_CHECK && NPW_IS_PLUGIN
+extern bool pid_check(void);
+#else
+#define pid_check() true
+#endif
+
 
 /* ====================================================================== */
 /* === NPClass Bridge                                                 === */
 /* ====================================================================== */
 
+static void g_NPClass_Invalidate(NPObject *npobj);
+static bool g_NPClass_HasMethod(NPObject *npobj, NPIdentifier name);
+static bool g_NPClass_Invoke(NPObject *npobj, NPIdentifier name, const NPVariant *args, uint32_t argCount, NPVariant *result);
+static bool g_NPClass_InvokeDefault(NPObject *npobj, const NPVariant *args, uint32_t argCount, NPVariant *result);
+static bool g_NPClass_HasProperty(NPObject *npobj, NPIdentifier name);
+static bool g_NPClass_GetProperty(NPObject *npobj, NPIdentifier name, NPVariant *result);
+static bool g_NPClass_SetProperty(NPObject *npobj, NPIdentifier name, const NPVariant *value);
+static bool g_NPClass_RemoveProperty(NPObject *npobj, NPIdentifier name);
+
 NPClass npclass_bridge = {
-  NP_CLASS_STRUCT_VERSION,
+  NPW_NP_CLASS_STRUCT_VERSION,
   NULL,
   NULL,
-  npclass_invoke_Invalidate,
-  npclass_invoke_HasMethod,
-  npclass_invoke_Invoke,
-  npclass_invoke_InvokeDefault,
-  npclass_invoke_HasProperty,
-  npclass_invoke_GetProperty,
-  npclass_invoke_SetProperty,
-  npclass_invoke_RemoveProperty
+  g_NPClass_Invalidate,
+  g_NPClass_HasMethod,
+  g_NPClass_Invoke,
+  g_NPClass_InvokeDefault,
+  g_NPClass_HasProperty,
+  g_NPClass_GetProperty,
+  g_NPClass_SetProperty,
+  g_NPClass_RemoveProperty
 };
 
+static inline bool is_valid_npobject_class(NPObject *npobj)
+{
+  if (npobj == NULL || npobj->_class == NULL)
+	return false;
+  NPObjectInfo *npobj_info = npobject_info_lookup(npobj);
+  if (npobj_info == NULL)
+	return false;
+  if (!npobj_info->is_valid)
+	npw_printf("ERROR: NPObject %p is no longer valid!\n", npobj);
+  return npobj_info->is_valid;
+}
+
 // NPClass::Invalidate
 int npclass_handle_Invalidate(rpc_connection_t *connection)
 {
+  D(bug("npclass_handle_Invalidate\n"));
+
   NPObject *npobj;
   int error = rpc_method_get_args(connection,
 								  RPC_TYPE_NP_OBJECT, &npobj,
@@ -72,17 +95,19 @@ int npclass_handle_Invalidate(rpc_connection_t *connection)
 	return error;
   }
 
-  if (npobj && npobj->_class && npobj->_class->invalidate) {
-	D(bug("NPClass::Invalidate(npobj %p)\n", npobj));
+  if (npobj && is_valid_npobject_class(npobj) && npobj->_class->invalidate) {
+	D(bugiI("NPClass::Invalidate(npobj %p)\n", npobj));
 	npobj->_class->invalidate(npobj);
-	D(bug(" done\n"));
+	D(bugiD("NPClass::Invalidate done\n"));
   }
 
-  return RPC_ERROR_NO_ERROR;
+  return rpc_method_send_reply(connection, RPC_TYPE_INVALID);
 }
 
-void npclass_invoke_Invalidate(NPObject *npobj)
+static void npclass_invoke_Invalidate(NPObject *npobj)
 {
+  npw_return_if_fail(rpc_method_invoke_possible(g_rpc_connection));
+
   int error = rpc_method_invoke(g_rpc_connection,
 								RPC_METHOD_NPCLASS_INVALIDATE,
 								RPC_TYPE_NP_OBJECT, npobj,
@@ -101,9 +126,26 @@ void npclass_invoke_Invalidate(NPObject *npobj)
   }
 }
 
+void g_NPClass_Invalidate(NPObject *npobj)
+{
+  if (!is_valid_npobject_class(npobj))
+	return;
+
+  if (!pid_check()) {
+	npw_printf("WARNING: NPClass::Invalidate called from the wrong process\n");
+	return;
+  }
+
+  D(bugiI("NPClass::Invalidate(npobj %p)\n", npobj));
+  npclass_invoke_Invalidate(npobj);
+  D(bugiD("NPClass::Invalidate done\n"));
+}
+
 // NPClass::HasMethod
 int npclass_handle_HasMethod(rpc_connection_t *connection)
 {
+  D(bug("npclass_handle_HasMethod\n"));
+
   NPObject *npobj;
   NPIdentifier name;
   int error = rpc_method_get_args(connection,
@@ -116,18 +158,22 @@ int npclass_handle_HasMethod(rpc_connection_t *connection)
 	return error;
   }
 
-  uint32_t ret = 0;
-  if (npobj && npobj->_class && npobj->_class->hasMethod) {
-	D(bug("NPClass::HasMethod(npobj %p, name id %p)\n", npobj, name));
+  uint32_t ret = false;
+  if (npobj && is_valid_npobject_class(npobj) && npobj->_class->hasMethod) {
+	D(bugiI("NPClass::HasMethod(npobj %p, name id %p)\n", npobj, name));
 	ret = npobj->_class->hasMethod(npobj, name);
-	D(bug(" return: %d\n", ret));
+	D(bugiD("NPClass::HasMethod return: %d\n", ret));
   }
 
-  return rpc_method_send_reply(connection, RPC_TYPE_UINT32, ret, RPC_TYPE_INVALID);
+  return rpc_method_send_reply(connection,
+							   RPC_TYPE_UINT32, ret,
+							   RPC_TYPE_INVALID);
 }
 
-bool npclass_invoke_HasMethod(NPObject *npobj, NPIdentifier name)
+static bool npclass_invoke_HasMethod(NPObject *npobj, NPIdentifier name)
 {
+  npw_return_val_if_fail(rpc_method_invoke_possible(g_rpc_connection), false);
+
   int error = rpc_method_invoke(g_rpc_connection,
 								RPC_METHOD_NPCLASS_HAS_METHOD,
 								RPC_TYPE_NP_OBJECT, npobj,
@@ -150,9 +196,27 @@ bool npclass_invoke_HasMethod(NPObject *npobj, NPIdentifier name)
   return ret;
 }
 
+bool g_NPClass_HasMethod(NPObject *npobj, NPIdentifier name)
+{
+  if (!is_valid_npobject_class(npobj))
+	return false;
+
+  if (!pid_check()) {
+	npw_printf("WARNING: NPClass::HasMethod called from the wrong process\n");
+	return false;
+  }
+
+  D(bugiI("NPClass::HasMethod(npobj %p, name id %p)\n", npobj, name));
+  bool ret = npclass_invoke_HasMethod(npobj, name);
+  D(bugiD("NPClass::HasMethod return: %d\n", ret));
+  return ret;
+}
+
 // NPClass::Invoke
 int npclass_handle_Invoke(rpc_connection_t *connection)
 {
+  D(bug("npclass_handle_Invoke\n"));
+
   NPObject *npobj;
   NPIdentifier name;
   uint32_t argCount;
@@ -168,29 +232,37 @@ int npclass_handle_Invoke(rpc_connection_t *connection)
 	return error;
   }
 
-  uint32_t ret = 0;
+  uint32_t ret = false;
   NPVariant result;
   VOID_TO_NPVARIANT(result);
-  if (npobj && npobj->_class && npobj->_class->invoke) {
-	D(bug("NPClass::Invoke(npobj %p, name id %p)\n", npobj, name));
+  if (npobj && is_valid_npobject_class(npobj) && npobj->_class->invoke) {
+	D(bugiI("NPClass::Invoke(npobj %p, name id %p)\n", npobj, name));
+	print_npvariant_args(args, argCount);
 	ret = npobj->_class->invoke(npobj, name, args, argCount, &result);
-	D(bug(" return: %d\n", ret));
+	gchar *result_str = string_of_NPVariant(&result);
+	D(bugiD("NPClass::Invoke return: %d (%s)\n", ret, result_str));
+	g_free(result_str);
   }
 
-  free(args);
+  int rpc_ret = rpc_method_send_reply(connection,
+									  RPC_TYPE_UINT32, ret,
+									  RPC_TYPE_NP_VARIANT, &result,
+									  RPC_TYPE_INVALID);
 
-  return rpc_method_send_reply(connection,
-							   RPC_TYPE_UINT32, ret,
-							   RPC_TYPE_NP_VARIANT, &result,
-							   RPC_TYPE_INVALID);
+  if (args) {
+	for (int i = 0; i < argCount; i++)
+	  NPN_ReleaseVariantValue(&args[i]);
+	free(args);
+  }
+
+  NPN_ReleaseVariantValue(&result);
+  return rpc_ret;
 }
 
-bool npclass_invoke_Invoke(NPObject *npobj, NPIdentifier name, const NPVariant *args, uint32_t argCount,
-						   NPVariant *result)
+static bool npclass_invoke_Invoke(NPObject *npobj, NPIdentifier name, const NPVariant *args, uint32_t argCount,
+								  NPVariant *result)
 {
-  if (result == NULL)
-	return false;
-  VOID_TO_NPVARIANT(*result);
+  npw_return_val_if_fail(rpc_method_invoke_possible(g_rpc_connection), false);
 
   int error = rpc_method_invoke(g_rpc_connection,
 								RPC_METHOD_NPCLASS_INVOKE,
@@ -218,9 +290,35 @@ bool npclass_invoke_Invoke(NPObject *npobj, NPIdentifier name, const NPVariant *
   return ret;
 }
 
+bool g_NPClass_Invoke(NPObject *npobj, NPIdentifier name, const NPVariant *args, uint32_t argCount,
+					  NPVariant *result)
+{
+  if (result == NULL)
+	return false;
+  VOID_TO_NPVARIANT(*result);
+
+  if (!is_valid_npobject_class(npobj))
+	return false;
+
+  if (!pid_check()) {
+	npw_printf("WARNING: NPClass::Invoke called from the wrong process\n");
+	return false;
+  }
+
+  D(bugiI("NPClass::Invoke(npobj %p, name id %p)\n", npobj, name));
+  print_npvariant_args(args, argCount);
+  bool ret = npclass_invoke_Invoke(npobj, name, args, argCount, result);
+  gchar *result_str = string_of_NPVariant(result);
+  D(bugiD("NPClass::Invoke return: %d (%s)\n", ret, result_str));
+  g_free(result_str);
+  return ret;
+}
+
 // NPClass::InvokeDefault
 int npclass_handle_InvokeDefault(rpc_connection_t *connection)
 {
+  D(bug("npclass_handle_InvokeDefault\n"));
+
   NPObject *npobj;
   uint32_t argCount;
   NPVariant *args;
@@ -234,29 +332,37 @@ int npclass_handle_InvokeDefault(rpc_connection_t *connection)
 	return error;
   }
 
-  uint32_t ret = 0;
+  uint32_t ret = false;
   NPVariant result;
   VOID_TO_NPVARIANT(result);
-  if (npobj && npobj->_class && npobj->_class->invokeDefault) {
-	D(bug("NPClass::InvokeDefault(npobj %p)\n", npobj));
+  if (npobj && is_valid_npobject_class(npobj) && npobj->_class->invokeDefault) {
+	D(bugiI("NPClass::InvokeDefault(npobj %p)\n", npobj));
+	print_npvariant_args(args, argCount);
 	ret = npobj->_class->invokeDefault(npobj, args, argCount, &result);
-	D(bug(" return: %d\n", ret));
+	gchar *result_str = string_of_NPVariant(&result);
+	D(bugiD("NPClass::InvokeDefault return: %d (%s)\n", ret, result_str));
+	g_free(result_str);
   }
 
-  free(args);
+  int rpc_ret = rpc_method_send_reply(connection,
+									  RPC_TYPE_UINT32, ret,
+									  RPC_TYPE_NP_VARIANT, &result,
+									  RPC_TYPE_INVALID);
 
-  return rpc_method_send_reply(connection,
-							   RPC_TYPE_UINT32, ret,
-							   RPC_TYPE_NP_VARIANT, &result,
-							   RPC_TYPE_INVALID);
+  if (args) {
+	for (int i = 0; i < argCount; i++)
+	  NPN_ReleaseVariantValue(&args[i]);
+	free(args);
+  }
+
+  NPN_ReleaseVariantValue(&result);
+  return rpc_ret;
 }
 
-bool npclass_invoke_InvokeDefault(NPObject *npobj, const NPVariant *args, uint32_t argCount,
-						   NPVariant *result)
+static bool npclass_invoke_InvokeDefault(NPObject *npobj, const NPVariant *args, uint32_t argCount,
+										 NPVariant *result)
 {
-  if (result == NULL)
-	return false;
-  VOID_TO_NPVARIANT(*result);
+  npw_return_val_if_fail(rpc_method_invoke_possible(g_rpc_connection), false);
 
   int error = rpc_method_invoke(g_rpc_connection,
 								RPC_METHOD_NPCLASS_INVOKE_DEFAULT,
@@ -283,9 +389,35 @@ bool npclass_invoke_InvokeDefault(NPObject *npobj, const NPVariant *args, uint32
   return ret;
 }
 
+bool g_NPClass_InvokeDefault(NPObject *npobj, const NPVariant *args, uint32_t argCount,
+							 NPVariant *result)
+{
+  if (result == NULL)
+	return false;
+  VOID_TO_NPVARIANT(*result);
+
+  if (!is_valid_npobject_class(npobj))
+	return false;
+
+  if (!pid_check()) {
+	npw_printf("WARNING: NPClass::InvokeDefault called from the wrong process\n");
+	return false;
+  }
+
+  D(bugiI("NPClass::InvokeDefault(npobj %p)\n", npobj));
+  print_npvariant_args(args, argCount);
+  bool ret = npclass_invoke_InvokeDefault(npobj, args, argCount, result);
+  gchar *result_str = string_of_NPVariant(result);
+  D(bugiD("NPClass::InvokeDefault return: %d (%s)\n", ret, result_str));
+  g_free(result_str);
+  return ret;
+}
+
 // NPClass::HasProperty
 int npclass_handle_HasProperty(rpc_connection_t *connection)
 {
+  D(bug("npclass_handle_HasProperty\n"));
+
   NPObject *npobj;
   NPIdentifier name;
   int error = rpc_method_get_args(connection,
@@ -298,18 +430,22 @@ int npclass_handle_HasProperty(rpc_connection_t *connection)
 	return error;
   }
 
-  uint32_t ret = 0;
-  if (npobj && npobj->_class && npobj->_class->hasProperty) {
-	D(bug("NPClass::HasProperty(npobj %p, name id %p)\n", npobj, name));
+  uint32_t ret = false;
+  if (npobj && is_valid_npobject_class(npobj) && npobj->_class->hasProperty) {
+	D(bugiI("NPClass::HasProperty(npobj %p, name id %p)\n", npobj, name));
 	ret = npobj->_class->hasProperty(npobj, name);
-	D(bug(" return: %d\n", ret));
+	D(bugiD("NPClass::HasProperty return: %d\n", ret));
   }
 
-  return rpc_method_send_reply(connection, RPC_TYPE_UINT32, ret, RPC_TYPE_INVALID);
+  return rpc_method_send_reply(connection,
+							   RPC_TYPE_UINT32, ret,
+							   RPC_TYPE_INVALID);
 }
 
-bool npclass_invoke_HasProperty(NPObject *npobj, NPIdentifier name)
+static bool npclass_invoke_HasProperty(NPObject *npobj, NPIdentifier name)
 {
+  npw_return_val_if_fail(rpc_method_invoke_possible(g_rpc_connection), false);
+
   int error = rpc_method_invoke(g_rpc_connection,
 								RPC_METHOD_NPCLASS_HAS_PROPERTY,
 								RPC_TYPE_NP_OBJECT, npobj,
@@ -332,9 +468,27 @@ bool npclass_invoke_HasProperty(NPObject *npobj, NPIdentifier name)
   return ret;
 }
 
+bool g_NPClass_HasProperty(NPObject *npobj, NPIdentifier name)
+{
+  if (!is_valid_npobject_class(npobj))
+	return false;
+
+  if (!pid_check()) {
+	npw_printf("WARNING: NPClass::HasProperty called from the wrong process\n");
+	return false;
+  }
+
+  D(bugiI("NPClass::HasProperty(npobj %p, name id %p)\n", npobj, name));
+  bool ret = npclass_invoke_HasProperty(npobj, name);
+  D(bugiD("NPClass::HasProperty return: %d\n", ret));
+  return ret;
+}
+  
 // NPClass::GetProperty
 int npclass_handle_GetProperty(rpc_connection_t *connection)
 {
+  D(bug("npclass_handle_GetProperty\n"));
+
   NPObject *npobj;
   NPIdentifier name;
   int error = rpc_method_get_args(connection,
@@ -347,26 +501,29 @@ int npclass_handle_GetProperty(rpc_connection_t *connection)
 	return error;
   }
 
-  uint32_t ret = 0;
+  uint32_t ret = false;
   NPVariant result;
   VOID_TO_NPVARIANT(result);
-  if (npobj && npobj->_class && npobj->_class->getProperty) {
-	D(bug("NPClass::GetProperty(npobj %p, name id %p)\n", npobj, name));
+  if (npobj && is_valid_npobject_class(npobj) && npobj->_class->getProperty) {
+	D(bugiI("NPClass::GetProperty(npobj %p, name id %p)\n", npobj, name));
 	ret = npobj->_class->getProperty(npobj, name, &result);
-	D(bug(" return: %d\n", ret));
+	gchar *result_str = string_of_NPVariant(&result);
+	D(bugiD("NPClass::GetProperty return: %d (%s)\n", ret, result_str));
+	g_free(result_str);
   }
 
-  return rpc_method_send_reply(connection,
-							   RPC_TYPE_UINT32, ret,
-							   RPC_TYPE_NP_VARIANT, &result,
-							   RPC_TYPE_INVALID);
+  int rpc_ret = rpc_method_send_reply(connection,
+									  RPC_TYPE_UINT32, ret,
+									  RPC_TYPE_NP_VARIANT, &result,
+									  RPC_TYPE_INVALID);
+
+  NPN_ReleaseVariantValue(&result);
+  return rpc_ret;
 }
 
-bool npclass_invoke_GetProperty(NPObject *npobj, NPIdentifier name, NPVariant *result)
+static bool npclass_invoke_GetProperty(NPObject *npobj, NPIdentifier name, NPVariant *result)
 {
-  if (result == NULL)
-	return false;
-  VOID_TO_NPVARIANT(*result);
+  npw_return_val_if_fail(rpc_method_invoke_possible(g_rpc_connection), false);
 
   int error = rpc_method_invoke(g_rpc_connection,
 								RPC_METHOD_NPCLASS_GET_PROPERTY,
@@ -393,9 +550,33 @@ bool npclass_invoke_GetProperty(NPObject *npobj, NPIdentifier name, NPVariant *r
   return ret;
 }
 
+bool g_NPClass_GetProperty(NPObject *npobj, NPIdentifier name, NPVariant *result)
+{
+  if (result == NULL)
+	return false;
+  VOID_TO_NPVARIANT(*result);
+
+  if (!is_valid_npobject_class(npobj))
+	return false;
+
+  if (!pid_check()) {
+	npw_printf("WARNING: NPClass::GetProperty called from the wrong process\n");
+	return false;
+  }
+
+  D(bugiI("NPClass::GetProperty(npobj %p, name id %p)\n", npobj, name));
+  bool ret = npclass_invoke_GetProperty(npobj, name, result);
+  gchar *result_str = string_of_NPVariant(result);
+  D(bugiD("NPClass::GetProperty return: %d (%s)\n", ret, result_str));
+  g_free(result_str);
+  return ret;
+}
+  
 // NPClass::SetProperty
 int npclass_handle_SetProperty(rpc_connection_t *connection)
 {
+  D(bug("npclass_handle_SetProperty\n"));
+
   NPObject *npobj;
   NPIdentifier name;
   NPVariant value;
@@ -410,22 +591,24 @@ int npclass_handle_SetProperty(rpc_connection_t *connection)
 	return error;
   }
 
-  uint32_t ret = 0;
-  if (npobj && npobj->_class && npobj->_class->setProperty) {
-	D(bug("NPClass::SetProperty(npobj %p, name id %p)\n", npobj, name));
+  uint32_t ret = false;
+  if (npobj && is_valid_npobject_class(npobj) && npobj->_class->setProperty) {
+	D(bugiI("NPClass::SetProperty(npobj %p, name id %p)\n", npobj, name));
 	ret = npobj->_class->setProperty(npobj, name, &value);
-	D(bug(" return: %d\n", ret));
+	D(bugiD("NPClass::SetProperty return: %d\n", ret));
   }
 
-  return rpc_method_send_reply(connection,
-							   RPC_TYPE_UINT32, ret,
-							   RPC_TYPE_INVALID);
+  int rpc_ret = rpc_method_send_reply(connection,
+									  RPC_TYPE_UINT32, ret,
+									  RPC_TYPE_INVALID);
+
+  NPN_ReleaseVariantValue(&value);
+  return rpc_ret;
 }
 
-bool npclass_invoke_SetProperty(NPObject *npobj, NPIdentifier name, const NPVariant *value)
+static bool npclass_invoke_SetProperty(NPObject *npobj, NPIdentifier name, const NPVariant *value)
 {
-  if (value == NULL)
-	return false;
+  npw_return_val_if_fail(rpc_method_invoke_possible(g_rpc_connection), false);
 
   int error = rpc_method_invoke(g_rpc_connection,
 								RPC_METHOD_NPCLASS_SET_PROPERTY,
@@ -452,9 +635,32 @@ bool npclass_invoke_SetProperty(NPObject *npobj, NPIdentifier name, const NPVari
   return ret;
 }
 
+bool g_NPClass_SetProperty(NPObject *npobj, NPIdentifier name, const NPVariant *value)
+{
+  if (value == NULL) {
+	npw_printf("WARNING: NPClass::SetProperty() called with a NULL value\n");
+	return false;
+  }
+
+  if (!is_valid_npobject_class(npobj))
+	return false;
+
+  if (!pid_check()) {
+	npw_printf("WARNING: NPClass::SetProperty called from the wrong process\n");
+	return false;
+  }
+
+  D(bugiI("NPClass::SetProperty(npobj %p, name id %p)\n", npobj, name));
+  bool ret = npclass_invoke_SetProperty(npobj, name, value);
+  D(bugiD("NPClass::SetProperty return: %d\n", ret));
+  return ret;
+}
+
 // NPClass::RemoveProperty
 int npclass_handle_RemoveProperty(rpc_connection_t *connection)
 {
+  D(bug("npclass_handle_RemoveProperty\n"));
+
   NPObject *npobj;
   NPIdentifier name;
   int error = rpc_method_get_args(connection,
@@ -467,18 +673,22 @@ int npclass_handle_RemoveProperty(rpc_connection_t *connection)
 	return error;
   }
 
-  uint32_t ret = 0;
-  if (npobj && npobj->_class && npobj->_class->removeProperty) {
-	D(bug("NPClass::RemoveProperty(npobj %p, name id %p)\n", npobj, name));
+  uint32_t ret = false;
+  if (npobj && is_valid_npobject_class(npobj) && npobj->_class->removeProperty) {
+	D(bugiI("NPClass::RemoveProperty(npobj %p, name id %p)\n", npobj, name));
 	ret = npobj->_class->removeProperty(npobj, name);
-	D(bug(" return: %d\n", ret));
+	D(bugiD("NPClass::RemoveProperty return: %d\n", ret));
   }
 
-  return rpc_method_send_reply(connection, RPC_TYPE_UINT32, ret, RPC_TYPE_INVALID);
+  return rpc_method_send_reply(connection,
+							   RPC_TYPE_UINT32, ret,
+							   RPC_TYPE_INVALID);
 }
 
-bool npclass_invoke_RemoveProperty(NPObject *npobj, NPIdentifier name)
+static bool npclass_invoke_RemoveProperty(NPObject *npobj, NPIdentifier name)
 {
+  npw_return_val_if_fail(rpc_method_invoke_possible(g_rpc_connection), false);
+
   int error = rpc_method_invoke(g_rpc_connection,
 								RPC_METHOD_NPCLASS_REMOVE_PROPERTY,
 								RPC_TYPE_NP_OBJECT, npobj,
@@ -501,6 +711,22 @@ bool npclass_invoke_RemoveProperty(NPObject *npobj, NPIdentifier name)
   return ret;
 }
 
+bool g_NPClass_RemoveProperty(NPObject *npobj, NPIdentifier name)
+{
+  if (!is_valid_npobject_class(npobj))
+	return false;
+
+  if (!pid_check()) {
+	npw_printf("WARNING: NPClass::RemoveProperty called from the wrong process\n");
+	return false;
+  }
+  
+  D(bugiI("NPClass::RemoveProperty(npobj %p, name id %p)\n", npobj, name));
+  bool ret = npclass_invoke_RemoveProperty(npobj, name);
+  D(bugiD("NPClass::RemoveProperty return: %d\n", ret));
+  return ret;
+}
+
 
 /* ====================================================================== */
 /* === NPObjectInfo                                                   === */
@@ -508,19 +734,22 @@ bool npclass_invoke_RemoveProperty(NPObject *npobj, NPIdentifier name)
 
 NPObjectInfo *npobject_info_new(NPObject *npobj)
 {
-  NPObjectInfo *npobj_info = malloc(sizeof(*npobj_info));
+  NPObjectInfo *npobj_info = NPW_MemNew0(NPObjectInfo, 1);
   if (npobj_info) {
 	static uint32_t id;
 	npobj_info->npobj = npobj;
 	npobj_info->npobj_id = ++id;
+	npobj_info->is_valid = true;
   }
   return npobj_info;
 }
 
 void npobject_info_destroy(NPObjectInfo *npobj_info)
 {
-  if (npobj_info)
-	free(npobj_info);
+  if (npobj_info) {
+	npw_plugin_instance_unref(npobj_info->plugin);
+	NPW_MemFree(npobj_info);
+  }
 }
 
 
@@ -567,6 +796,7 @@ NPObject *npobject_new(uint32_t npobj_id, NPP instance, NPClass *class)
 	return NULL;
   }
   npobj_info->npobj_id = npobj_id;
+  npobj_info->plugin = npw_plugin_instance_ref(NPW_PLUGIN_INSTANCE(instance));
   npobject_associate(npobj, npobj_info);
   return npobj;
 }
@@ -581,7 +811,7 @@ void npobject_destroy(NPObject *npobj)
 
 void npobject_associate(NPObject *npobj, NPObjectInfo *npobj_info)
 {
-  assert(npobj && npobj_info && npobj_info > 0);
+  assert(npobj && npobj_info && npobj_info->npobj_id > 0);
   npobject_hash_table_insert(npobj, npobj_info);
 }
 
@@ -605,10 +835,14 @@ bool npobject_bridge_new(void)
 
 void npobject_bridge_destroy(void)
 {
-  if (g_npobject_ids)
+  if (g_npobject_ids) {
 	g_hash_table_destroy(g_npobject_ids);
-  if (g_npobjects)
+	g_npobject_ids = NULL;
+  }
+  if (g_npobjects) {
 	g_hash_table_destroy(g_npobjects);
+	g_npobjects = NULL;
+  }
 }
 
 void npobject_hash_table_insert(NPObject *npobj, NPObjectInfo *npobj_info)
@@ -638,3 +872,108 @@ NPObject *npobject_lookup(uint32_t npobj_id)
 {
   return g_hash_table_lookup(g_npobject_ids, (void *)(uintptr_t)npobj_id);
 }
+
+static void npruntime_deactivate_func(gpointer key, gpointer value, gpointer user_data)
+{
+  NPObjectInfo *npobj_info = (NPObjectInfo *)value;
+  npobj_info->is_valid = false;
+}
+
+void npruntime_deactivate(void)
+{
+  g_hash_table_foreach(g_npobjects, npruntime_deactivate_func, NULL);
+}
+
+
+/* ====================================================================== */
+/* === NPVariant helpers                                              === */
+/* ====================================================================== */
+
+void
+npvariant_clear(NPVariant *variant)
+{
+  switch (variant->type) {
+  case NPVariantType_Void:
+  case NPVariantType_Null:
+  case NPVariantType_Bool:
+  case NPVariantType_Int32:
+  case NPVariantType_Double:
+	break;
+  case NPVariantType_String:
+	{
+	  NPString *s = &NPVARIANT_TO_STRING(*variant);
+	  if (s->utf8characters)
+		NPN_MemFree((void *)s->utf8characters);
+	  break;
+	}
+  case NPVariantType_Object:
+	{
+	  NPObject *npobj = NPVARIANT_TO_OBJECT(*variant);
+	  if (npobj)
+		NPN_ReleaseObject(npobj);
+	  break;
+	}
+  }
+  VOID_TO_NPVARIANT(*variant);
+}
+
+// Make sure to deallocate with g_free() since it comes from a GString
+gchar *
+string_of_NPVariant(const NPVariant *arg)
+{
+#if DEBUG
+  if (arg == NULL)
+	return NULL;
+  GString *str = g_string_new(NULL);
+  switch (arg->type)
+	{
+	case NPVariantType_Void:
+	  g_string_append_printf(str, "void");
+	  break;
+	case NPVariantType_Null:
+	  g_string_append_printf(str, "null");
+	  break;
+	case NPVariantType_Bool:
+	  g_string_append(str, arg->value.boolValue ? "true" : "false");
+	  break;
+	case NPVariantType_Int32:
+	  g_string_append_printf(str, "%d", arg->value.intValue);
+	  break;
+	case NPVariantType_Double:
+	  g_string_append_printf(str, "%f", arg->value.doubleValue);
+	  break;
+	case NPVariantType_String:
+	  g_string_append_c(str, '\'');
+	  g_string_append_len(str,
+						  arg->value.stringValue.utf8characters,
+						  arg->value.stringValue.utf8length);
+	  g_string_append_c(str, '\'');
+	  break;
+	case NPVariantType_Object:
+	  g_string_append_printf(str, "<object %p>", arg->value.objectValue);
+	  break;
+	default:
+	  g_string_append_printf(str, "<invalid type %d>", arg->type);
+	  break;
+	}
+  return g_string_free(str, FALSE);
+#endif
+  return NULL;
+}
+
+void
+print_npvariant_args(const NPVariant *args, uint32_t nargs)
+{
+#if DEBUG
+  GString *str = g_string_new(NULL);
+  for (int i = 0; i < nargs; i++) {
+	if (i > 0)
+	  g_string_append(str, ", ");
+	gchar *argstr = string_of_NPVariant(&args[i]);
+	g_string_append(str, argstr);
+	g_free(argstr);
+  }
+  D(bug("%u args (%s)\n", nargs, str->str));
+  g_string_free(str, TRUE);
+#endif
+}
