@@ -1,7 +1,7 @@
 /*
  *  npw-config.c - nspluginwrapper configuration tool
  *
- *  nspluginwrapper (C) 2005-2007 Gwenole Beauchesne
+ *  nspluginwrapper (C) 2005-2009 Gwenole Beauchesne
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -13,9 +13,9 @@
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *  You should have received a copy of the GNU General Public License along
+ *  with this program; if not, write to the Free Software Foundation, Inc.,
+ *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 #include "sysdeps.h"
@@ -41,6 +41,7 @@
 
 static bool g_auto = false;
 static bool g_verbose = false;
+static bool g_allow_native = false;
 static const char NPW_CONFIG[] = "nspluginwrapper";
 
 static void error(const char *format, ...)
@@ -140,6 +141,14 @@ static const char *get_system_mozilla_plugin_dir(void)
 	  };
 	  dirs = netbsd_dirs;
 	}
+#elif defined(__sun__)
+	{
+	  static const char *solaris_dirs[] = {
+		LIBDIR "/firefox/plugins",
+		"/usr/sfw/" LIB "/mozilla/plugins",
+	  };
+	  dirs = solaris_dirs;
+	}
 #elif defined(__linux__)
 	if (access("/etc/SuSE-release", F_OK) == 0) {
 	  static const char *suse_dirs[] = {
@@ -227,6 +236,9 @@ static const char **get_mozilla_plugin_dirs(void)
 	"/usr/pkg/lib/RealPlayer/mozilla",
 	"/usr/pkg/Acrobat5/Browsers/intellinux",
 	"/usr/pkg/Acrobat7/Browser/intellinux",
+#endif
+#if defined(__sun__)
+	"/usr/sfw/lib/mozilla/plugins",
 #endif
   };
 
@@ -504,7 +516,8 @@ static int detect_plugin_viewer(const char *filename, NPW_PluginInfo *out_plugin
 	target_os_table[0] = NULL;
 
   // don't wrap plugins for host OS/ARCH
-  if (out_plugin_info
+  if (!g_allow_native
+	  && out_plugin_info
 	  && out_plugin_info->target_arch && strcmp(out_plugin_info->target_arch, HOST_ARCH) == 0
 	  && out_plugin_info->target_os && strcmp(out_plugin_info->target_os, HOST_OS) == 0)
 	return EXIT_VIEWER_NATIVE;
@@ -531,6 +544,11 @@ static int detect_plugin_viewer(const char *filename, NPW_PluginInfo *out_plugin
 	  if (pid < 0)
 		continue;
 	  else if (pid == 0) {
+		if (!g_verbose) {
+		  // don't spit out errors in non-verbose mode, we only need
+		  // to know whether there is a valid viewer or not
+		  freopen("/dev/null", "w", stderr);
+		}
 		execl(viewer_path, NPW_VIEWER, "--test", "--plugin", filename, NULL);
 		exit(1);
 	  }
@@ -618,7 +636,7 @@ static bool is_wrapper_plugin_0(const char *plugin_path)
 	&& strcmp(plugin_info.path, NPW_OLD_DEFAULT_PLUGIN_PATH) != 0;		// exclude ARCH npwrapper.so
 }
 
-static bool is_system_wide_wrapper_plugin(const char *plugin_path)
+static bool has_system_wide_wrapper_plugin(const char *plugin_path, bool check_ident)
 {
   char *plugin_base = strrchr(plugin_path, '/');
   if (plugin_base == NULL)
@@ -637,10 +655,41 @@ static bool is_system_wide_wrapper_plugin(const char *plugin_path)
   NPW_PluginInfo plugin_info;
   return (is_wrapper_plugin(s_plugin_path, &plugin_info)
 		  && strcmp(plugin_info.path, plugin_path) == 0
-		  && strcmp(plugin_info.ident, NPW_PLUGIN_IDENT) == 0
+		  && (check_ident ?
+			  strcmp(plugin_info.ident, NPW_PLUGIN_IDENT) == 0 :
+			  true)
 		  && plugin_info.mtime == st.st_mtime);
 }
 
+static bool match_path_prefix(const char *path, const char *prefix)
+{
+  if (path == NULL || prefix == NULL)
+	return false;
+  int prefix_len = strlen(prefix);
+  return strncmp(path, prefix, prefix_len) == 0 && path[prefix_len] == '/';
+}
+
+static bool is_user_home_path(const char *path)
+{
+  const char *homedir;
+  if ((homedir = get_user_home_dir()) == NULL)
+	return false;
+  return match_path_prefix(path, homedir);
+}
+
+static bool is_root_path(const char *path)
+{
+  struct passwd *pwent = getpwuid(0);
+  return pwent && match_path_prefix(path, pwent->pw_dir);
+}
+
+static bool is_root_only_accessible_plugin(const char *plugin_path)
+{
+  /* XXX: this is very primitive and doesn't account for non ~root/
+	 directories that actually are not accessible by non root users */
+  return plugin_path && is_root_path(plugin_path);
+}
+
 typedef bool (*is_plugin_cb)(const char *plugin_path, NPW_PluginInfo *plugin_info);
 typedef int (*process_plugin_cb)(const char *plugin_path, NPW_PluginInfo *plugin_info);
 
@@ -690,10 +739,6 @@ static int do_install_plugin(const char *plugin_path, const char *plugin_dir, NP
   if (n < 0 || n >= sizeof(d_plugin_path))
 	return 3;
 
-  int mode = 0700;
-  if (geteuid() == 0 && strcmp(plugin_dir, "/root") != 0)
-	mode = 0755;
-
   NPW_PluginInfo w_plugin_info;
   if (!is_wrapper_plugin(NPW_DEFAULT_PLUGIN_PATH, &w_plugin_info))
 	return 5;
@@ -750,6 +795,11 @@ static int do_install_plugin(const char *plugin_path, const char *plugin_dir, NP
   strcpy(pi->target_arch, plugin_info->target_arch);
   strcpy(pi->target_os, plugin_info->target_os);
 
+  int mode = 0700;
+  if (!is_user_home_path(d_plugin_path) &&
+	  !is_root_only_accessible_plugin(plugin_dir))
+	mode = 0755;
+
   int d_fd = open(d_plugin_path, O_CREAT | O_WRONLY, mode);
   if (d_fd < 0)
 	return 4;
@@ -772,12 +822,15 @@ static int install_plugin(const char *plugin_path, NPW_PluginInfo *plugin_info)
   if (g_verbose)
 	printf("Install plugin %s\n", plugin_path);
 
-  ret = do_install_plugin(plugin_path, get_system_mozilla_plugin_dir(), plugin_info);
-  if (ret == 0)
-	return 0;
+  // don't install plugin system-wide if it is only accessible by root
+  if (!is_root_only_accessible_plugin(plugin_path)) {
+	ret = do_install_plugin(plugin_path, get_system_mozilla_plugin_dir(), plugin_info);
+	if (ret == 0)
+	  return 0;
+  }
 
   // don't install plugin in user home dir if already available system-wide
-  if (is_system_wide_wrapper_plugin(plugin_path)) {
+  if (has_system_wide_wrapper_plugin(plugin_path, true)) {
 	if (g_verbose)
 	  printf(" ... already installed system-wide, skipping\n");
 	return 0;
@@ -810,7 +863,7 @@ static int auto_install_plugins(void)
   return 0;
 }
 
-static int remove_plugin(const char *plugin_path, ...)
+static int remove_plugin(const char *plugin_path)
 {
   if (g_verbose)
 	printf("Remove plugin %s\n", plugin_path);
@@ -821,6 +874,11 @@ static int remove_plugin(const char *plugin_path, ...)
   return 0;
 }
 
+static int remove_plugin_cb(const char *plugin_path, void *unused)
+{
+  return remove_plugin(plugin_path);
+}
+
 static int auto_remove_plugins(void)
 {
   const char **plugin_dirs = get_mozilla_plugin_dirs();
@@ -830,14 +888,14 @@ static int auto_remove_plugins(void)
 	  const char *plugin_dir = plugin_dirs[i];
 	  if (g_verbose)
 		printf("Auto-remove plugins from %s\n", plugin_dir);
-	  process_plugin_dir(plugin_dir, (is_plugin_cb)is_wrapper_plugin_0, (process_plugin_cb)remove_plugin);
+	  process_plugin_dir(plugin_dir, (is_plugin_cb)is_wrapper_plugin_0, (process_plugin_cb)remove_plugin_cb);
 	}
   }
   free(plugin_dirs);
   return 0;
 }
 
-static int update_plugin(const char *plugin_path, ...)
+static int update_plugin(const char *plugin_path)
 {
   if (g_verbose)
 	printf("Update plugin %s\n", plugin_path);
@@ -850,18 +908,31 @@ static int update_plugin(const char *plugin_path, ...)
 
   if (access(plugin_info.path, F_OK) < 0) {
 	if (g_verbose)
-	  printf("  NS4 plugin %s is no longer available, removing wrapper\n", plugin_info.path);
+	  printf("  NPAPI plugin %s is no longer available, removing wrapper\n", plugin_info.path);
 	ret = remove_plugin(plugin_path);
   }
-  else if (is_system_wide_wrapper_plugin(plugin_info.path)
+  else if (has_system_wide_wrapper_plugin(plugin_info.path, true)
 		   && !strstart(plugin_path, get_system_mozilla_plugin_dir(), NULL)) {
 	if (g_verbose)
-	  printf("  NS4 plugin %s is already installed system-wide, removing wrapper\n", plugin_info.path);
+	  printf("  NPAPI plugin %s is already installed system-wide, removing wrapper\n", plugin_info.path);
 	ret = remove_plugin(plugin_path);
   }
+  else if (has_system_wide_wrapper_plugin(plugin_info.path, false)
+		   && is_root_only_accessible_plugin(plugin_info.path)) {
+	/* Don't check for an exact ident, we only need to know if there
+	   is a system-wide plugin to be removed. It will then be
+	   re-installed to the user (root) private mozilla plugins dir */
+	if (g_verbose)
+	  printf("  NPAPI plugin %s is accessible by root only, removing wrapper\n", plugin_info.path);
+	if ((ret = remove_plugin(plugin_path)) != 0)
+	  return ret;
+	if (g_verbose)
+	  printf ("  ... but re-installing it to root private mozilla plugins dir\n");
+	ret = install_plugin(plugin_info.path, &plugin_info);
+  }
   else if (stat(plugin_info.path, &st) == 0 && st.st_mtime > plugin_info.mtime) {
 	if (g_verbose)
-	  printf("  NS4 plugin %s was modified, reinstalling plugin\n", plugin_info.path);
+	  printf("  NPAPI plugin %s was modified, reinstalling plugin\n", plugin_info.path);
 	ret = install_plugin(plugin_info.path, &plugin_info);
   }
   else if (strcmp(plugin_info.ident, NPW_PLUGIN_IDENT) != 0) {
@@ -873,6 +944,11 @@ static int update_plugin(const char *plugin_path, ...)
   return ret;
 }
 
+static int update_plugin_cb(const char *plugin_path, void *unused)
+{
+  return update_plugin(plugin_path);
+}
+
 static int auto_update_plugins(void)
 {
   const char **plugin_dirs = get_mozilla_plugin_dirs();
@@ -882,14 +958,14 @@ static int auto_update_plugins(void)
 	  const char *plugin_dir = plugin_dirs[i];
 	  if (g_verbose)
 		printf("Auto-update plugins from %s\n", plugin_dir);
-	  process_plugin_dir(plugin_dir, (is_plugin_cb)is_wrapper_plugin_0, (process_plugin_cb)update_plugin);
+	  process_plugin_dir(plugin_dir, (is_plugin_cb)is_wrapper_plugin_0, (process_plugin_cb)update_plugin_cb);
 	}
   }
   free(plugin_dirs);
   return 0;
 }
 
-static int list_plugin(const char *plugin_path, ...)
+static int list_plugin(const char *plugin_path)
 {
   NPW_PluginInfo plugin_info;
   is_wrapper_plugin(plugin_path, &plugin_info);
@@ -911,6 +987,11 @@ static int list_plugin(const char *plugin_path, ...)
   return 0;
 }
 
+static int list_plugin_cb(const char *plugin_path, void *unused)
+{
+  return list_plugin(plugin_path);
+}
+
 static void print_usage(void)
 {
   printf("%s, configuration tool.  Version %s\n", NPW_CONFIG, NPW_VERSION);
@@ -920,6 +1001,7 @@ static void print_usage(void)
   printf("   -h --help               print this message\n");
   printf("   -v --verbose            flag: set verbose mode\n");
   printf("   -a --auto               flag: set automatic mode for plugins discovery\n");
+  printf("   -n --native             flag: allow native plugin(s) to be wrapped\n");
   printf("   -l --list               list plugins currently installed\n");
   printf("   -u --update             update plugin(s) currently installed\n");
   printf("   -i --install [FILE(S)]  install plugin(s)\n");
@@ -945,6 +1027,12 @@ static int process_auto(int argc, char *argv[])
   return 0;
 }
 
+static int process_native(int argc, char *argv[])
+{
+  g_allow_native = true;
+  return 0;
+}
+
 static int process_list(int argvc, char *argv[])
 {
   const char **plugin_dirs = get_mozilla_plugin_dirs();
@@ -954,7 +1042,7 @@ static int process_list(int argvc, char *argv[])
 	  const char *plugin_dir = plugin_dirs[i];
 	  if (g_verbose)
 		printf("List plugins in %s\n", plugin_dir);
-	  process_plugin_dir(plugin_dir, (is_plugin_cb)is_wrapper_plugin_0, (process_plugin_cb)list_plugin);
+	  process_plugin_dir(plugin_dir, (is_plugin_cb)is_wrapper_plugin_0, (process_plugin_cb)list_plugin_cb);
 	}
   }
   free(plugin_dirs);
@@ -1062,6 +1150,7 @@ int main(int argc, char *argv[])
 	{ 'h', "help",		process_help,		1 },
 	{ 'v', "verbose",	process_verbose,	0 },
 	{ 'a', "auto",		process_auto,		0 },
+	{ 'n', "native",	process_native,		0 },
 	{ 'l', "list",		process_list,		1 },
 	{ 'u', "update",	process_update,		1 },
 	{ 'i', "install",	process_install,	1 },
